<div aria-label="Pages"{{ paginator.attrs }}>
    <{{ paginator.container.tag }}{{ paginator.container.attrs }}>
        {% if show_first %}
            <{{ paginator.item.tag }}{{ paginator.item.attrs }}>
                <a href="?{{ extra|escape }}{{ paginator.iommi_path }}=1" aria-label="First Page"{{ paginator.link.attrs }}>&laquo;</a>
            </{{ paginator.item.tag }}>
        {% endif %}

        {% if has_previous %}
            <{{ paginator.item.tag }}{{ paginator.item.attrs }}>
                <a href="?{{ extra|escape }}{{ paginator.iommi_path }}={{ previous|stringformat:'s' }}" aria-label="Previous Page"{{ paginator.link.attrs }}>&lt;</a>
            </{{ paginator.item.tag }}>
        {% endif %}

        {% for num in page_numbers %}
            <{{ paginator.active_item.tag }}{% if num == page %}{{ paginator.active_item.attrs }}{% else %}{{ paginator.item.attrs }}{% endif %}>
                <a href="?{{ extra|escape }}{{ paginator.iommi_path }}={{ num|stringformat:'s' }}" aria-label="Page {{ num }}"{{ paginator.link.attrs }}>{{ num }}</a>
            </{{ paginator.active_item.tag }}>
        {% endfor %}

        {% if has_next %}
            <{{ paginator.item.tag }}{{ paginator.item.attrs }}>
                <a href="?{{ extra|escape }}{{ paginator.iommi_path }}={{ next|stringformat:'s' }}" aria-label="Next Page"{{ paginator.link.attrs }}>&gt;</a>
            </{{ paginator.item.tag }}>
        {% endif %}

        {% if show_last %}
            <{{ paginator.item.tag }}{{ paginator.item.attrs }}>
                <a href="?{{ extra|escape }}{{ paginator.iommi_path }}={{ pages|stringformat:'s' }}" aria-label="Last Page"{{ paginator.link.attrs }}>&raquo;</a>
            </{{ paginator.item.tag }}>
        {% endif %}
    </{{ paginator.container.tag }}>
</div>

