{% if not table.query.form.is_valid and table.invalid_form_message %}
    {{ table.invalid_form_message }}
{% elif not table.paginator.count and table.empty_message != None %}
    {{ table.empty_message }}
{% else %}

    <div class="iommi-table-plus-paginator">
        {{ table.table_tag_wrapper.iommi_open_tag }}
        {{ table.iommi_open_tag }}

            {{ table.header }}

            {{ table.tbody }}

        {{ table.iommi_close_tag }}
        {{ table.table_tag_wrapper.iommi_close_tag }}

        {{ table.paginator }}
    </div>

{% endif %}
